from gaelib.db import model, properties

class Test(model.Model):
    """
        Test database
    """
    string1 = properties.StringProperty()
    string2 = properties.StringProperty()
    int1 = properties.IntegerProperty()
    float1 = properties.FloatProperty()
    bool1 = properties.BooleanProperty()



from gaelib.tests.base import BaseUnitTestCase
from gaelib.storage import helpers


class HelpersTestCase(BaseUnitTestCase):

  def setUp(self):
    super().setUp()

  def test_get_default_storage_bucket(self):
    self.assertEqual(helpers.get_default_storage_bucket(),
                     'codejedi-crypticcup-staging.appspot.com')

  def test_get_storage_bucket(self):
    self.assertEqual(helpers.get_storage_bucket(),
                     'codejedi-crypticcup-staging.appspot.com')



import sys
import requests
import google.oauth2.id_token

from flask import g, request, session
from google.auth.transport import requests as grequests
from gaelib.env import get_dashboard_url_prefix, get_app_or_default_prop
from jose import jwt



def configure_login_callback(callback):
  setattr(sys.modules[__name__], 'login_callback', callback)

def get_user_id_and_token():
  auth_creds = request.authorization

  if auth_creds:
    user_id = auth_creds.username
    id_token = auth_creds.password
  else:
    user_id = session.get('gae_uid')
    id_token = session.get('gae_token')
  auth_type = ''

  try:
    auth_type = request.json.get('auth_type')
  except (AttributeError, KeyError):
    auth_type = ''

  if not auth_type:
    try:
      auth_type = request.form.get('auth_type')
    except (AttributeError, KeyError):
      auth_type = ''

  if not auth_type:
    try:
      auth_type = request.args.get('auth_type')
    except (AttributeError, KeyError):
      auth_type = ''

  if not auth_type:
    try:
      auth_type = request.params.get('auth_type')
    except (AttributeError, KeyError):
      auth_type = ''

  if not auth_type:
    auth_type = 'firebase'

  url = request.path
  is_dashboard_url = url.strip('/').startswith(get_dashboard_url_prefix())

  # Overriding id_token if firebaseaccesstoken is found in the browser cookie
  if is_dashboard_url and 'firebaseAccessToken' in request.cookies:
    id_token = str(request.cookies.get('firebaseAccessToken'))

  return user_id, id_token, auth_type, is_dashboard_url


class Auth(object):
  """
      Class that implements methods to authenticate requests.
  """
  firebase_request_adapter = grequests.Request()

  def __init__(self, id_token, user_id=None):
    self.user_id = user_id
    self.id_token = id_token
    self.logger = g.app.logger

  def firebase_authorize(self):
    '''
        Authorize id_token with firebase,
        and return associated claims.
    '''
    self.logger.info("Authorizing ID token with Firebase")
    claims = google.oauth2.id_token.verify_firebase_token(
        self.id_token,
        Auth.firebase_request_adapter
    )
    return claims


  def auth0_authorize(self):
    '''
        Authorize id_token with firebase,
        and return associated claims.
    '''
    token = self.id_token
    jwks = requests.get(get_app_or_default_prop('AUTH0_JKWS_DOMAIN')).json
    unverified_header = jwt.get_unverified_header(token)
    rsa_key = {}
    for key in jwks["keys"]:
      if key["kid"] == unverified_header["kid"]:
          rsa_key = {
              "kty": key["kty"],
              "kid": key["kid"],
              "use": key["use"],
              "n": key["n"],
              "e": key["e"]
          }
    if rsa_key:
      try:
          payload = jwt.decode(
              token,
              rsa_key,
              algorithms=["RS256"],
              audience=get_app_or_default_prop('AUTH0_API_AUDIENCE'),
              issuer="https://"+get_app_or_default_prop('AUTH0_DOMAIN')+"/"
          )
      except jwt.ExpiredSignatureError:
        self.logger.error('Expired Signature for %s', self.id_token)
        payload = {}
      except jwt.JWTClaimsError:
        self.logger.error('Wrong claims for %s', self.id_token)
        payload = {}
      except Exception as e:
        self.logger.error('Other Exception %s for %s', e, self.id_token)
        payload = {}

      self.logger.info("Auth0 payload received %s" % payload)
      return payload
    return {}

  def authorize_login_request(self, auth_type='firebase'):
    self.logger.info('Authorizing login request')
    if auth_type == 'firebase':
      claims = self.firebase_authorize()
    elif auth_type == 'auth0':
      self.logger.info('Auth0 login request for %s', self.id_token)
      claims = self.auth0_authorize()
    if claims:
      self.user_id = claims['sub']

    return claims

  def authorize_request(self, auth_type='firebase'):
    self.logger.info("Authorizing request")

    if auth_type == 'firebase':
      claims = self.firebase_authorize()
    elif auth_type == 'auth0':
      self.logger.info('Auth0 login request for %s', self.id_token)
      claims = self.auth0_authorize()

    if claims is None:
      return {}
    else:
      claimed_user_id = claims['sub']

    if claimed_user_id == self.user_id:
      return claims

    # claimed_user_id doesn't match request user_id
    return {}



from flask import g,render_template, request, session
from gaelib.auth.decorators import auth_required, access_control
from gaelib.auth.models import User, UserRole
from gaelib.env import get_dashboard_url_prefix, get_env, get_post_login_page, get_sidebar_template
from gaelib.view.base_view import BaseHttpHandler



class DashboardView(BaseHttpHandler):
  decorators = [access_control(role=UserRole.STAFF), auth_required]
  methods = ['GET', 'POST']

class DashboardLogin(BaseHttpHandler):
  """
      The solo view function to get all the game entities
      matching the filters passed
  """
  methods = ['GET']
  def dispatch_request(self):
    return render_template("login.html",
      post_login_page=get_post_login_page(),
      )


class DashboardLogout(BaseHttpHandler):
  """
      The solo view function to get all the game entities
      matching the filters passed
  """
  methods = ['GET']

  def dispatch_request(self):
    try:
      session.pop("gae_uid")
    except KeyError:
      pass
    return render_template("logout.html")


class Notifications(DashboardView):
  """
      The solo view function to get all the game entities
      matching the filters passed
  """
  methods = ['GET', 'POST']

  def dispatch_request(self):
    if request.method == 'GET':
      return render_template("notifications.html")





"""
    This module defines the property classes for attributes.
"""
from datetime import datetime
from google.cloud.datastore.key import Key


class Property():
  """
      The base property class for defining attribute types.
  """
  _repeated = None
  _default = None
  _choices = None

  def __init__(self, value, prop, repeated=None, default=None, choices=None):
    if repeated is not None:
      self._repeated = repeated
    if default is not None:
      self._default = default
    if choices is not None:
      self._choices = choices

    self.property = prop
    self.value = self.validate(value)

  def validate(self, value):
    """
        Method to validate the property value.
    """
    valid = True
    if self._repeated is True:
      if isinstance(value, list):
        for v in value:
          valid = valid and isinstance(v, self.property)

      if not valid:
        error_subject = "One repeated value"

    elif value and not isinstance(value, self.property):
      valid = False
      error_subject = "Value"

    if not value and self._default:
      value = self._default
      if not isinstance(value, self.property):
        valid = False
        error_subject = "Default value"

    choices = self._choices
    default = self._default
    if choices and value and value not in choices:
      raise ValueError("Value does not conform to the choices %s" % choices)

    if choices and default and default not in choices:
      raise ValueError(
          "Default Value does not conform to the choices %s" % choices)

    if not valid:
      raise ValueError("{} {} doesn't match the class type for {}".format(
          error_subject, str(value), self.property))

    return value


class StringProperty(Property):
  """
      Attribute property class for string type
  """

  def __init__(self, val=None, choices=None, default=None):
    super().__init__(val, str, default=default, choices=choices)


class FloatProperty(Property):
  """
      Attribute property class for float type
  """

  def __init__(self, value=None, repeated=None):
    super().__init__(value, float, repeated)


class IntegerProperty(Property):
  """
      Attribute property class for int type
  """

  def __init__(self, value=None, repeated=None, default=None, choices=None):
    # choices is used when you want to have an enum
    super().__init__(value, int, repeated, default=default, choices=choices)


class BooleanProperty(Property):
  """
      Attribute property class for int type
  """

  def __init__(self, value=None, repeated=None):
    super().__init__(value, bool, repeated)


class DateTimeProperty(Property):
  """
      Attribute property class for datetime type
  """

  def __init__(self, value=None, repeated=None):
    super().__init__(value, datetime, repeated)


class ReferenceProperty(Property):
  """
      Attribute property class for datastore key type.
  """

  def __init__(self,  value=None, reference_class=None, repeated=None):
    self.reference_class = reference_class
    super().__init__(value, Key, repeated)



import os
import logging

import jinja2
import constants
from flask import g, Flask, request
from google.cloud.logging.client import Client
from google.cloud.logging.handlers.app_engine import AppEngineHandler
from gaelib.env import is_dev, get_env, get_dashboard_url_prefix, get_sidebar_template, get_app_or_default_prop
from gaelib import filters
from gaelib.urls import auth_urls, dashboard_lib_urls, dashboard_lib_template_dir,client_logger_urls

PARAMETER_LOGGING = get_app_or_default_prop('PARAMETER_LOGGING')

app_template_dir = os.path.abspath('./templates/')

app = Flask(__name__, template_folder=app_template_dir)

# Uncomment to debug template loading issues
# app.config['EXPLAIN_TEMPLATE_LOADING'] = True

app.jinja_loader.searchpath.append(dashboard_lib_template_dir)


@app.before_request
def log_request_info():
  """
      Logs request params before dispatching request
  """
  g.app = app

  if not PARAMETER_LOGGING:
    return

  request_data = None
  request_args = request.args.to_dict()
  request_form = request.form.to_dict()
  request_json = request.json

  if request_args:
    g.app.logger.info('Request args: ' + str(request_args))
  if request_form:
    g.app.logger.info('Request form: ' + str(request_form))
  if request_json:
    g.app.logger.info('Request json: ' + str(request_json))

@app.context_processor
def inject_global_template_vars():
  return dict(app_name=get_app_or_default_prop('APP_NAME'), dashboard_prefix=get_dashboard_url_prefix(),
              env=get_env(), sidebar_template=get_sidebar_template())

def startup(auth=True, parameter_logging=False, client_logging=False, dashboard=True):
  """The startup script to create flask app and check if
      the application is running on local server or production.
  """

  if is_dev():
    os.environ['DATASTORE_EMULATOR_HOST'] = 'localhost:8089'
    os.environ['DATASTORE_PROJECT_ID'] = 'codejedi-crypticcup-staging'
    os.environ['GOOGLE_CLOUD_PROJECT'] = 'codejedi-crypticcup-staging'
    # Dumb hack we need to run locally
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = os.getcwd() + \
        '/../fake_creds.json'

  gcp_client = Client(os.environ['GOOGLE_CLOUD_PROJECT'])
  gcph = AppEngineHandler(gcp_client)
  app.logger.setLevel(logging.DEBUG)
  app.logger.addHandler(gcph)

  if auth:
    app.register_blueprint(auth_urls)

  if dashboard:
    app.register_blueprint(dashboard_lib_urls)

  app.register_blueprint(filters.blueprint)
  
  app.secret_key = get_app_or_default_prop('SESSION_SECRET')   # Used for session management

  if client_logging:
    app.register_blueprint(client_logger_urls)


  return app


from gaelib.view.base_view import BaseHttpHandler
from gaelib.auth.models import User, UserRole
from gaelib.auth import auth
from gaelib.env import get_app_or_default_prop
from flask import g, request, session



# pragma pylint: disable=too-few-public-methods


class Login(BaseHttpHandler):
  """
      The Login view function to sign in a user based on id_token.
      Creates a new User in datastore if user doesn't exist already.
  """

  def dispatch_request(self):
    """
        Process and dispatch response for request to Base-URL/login/
    """
    g.app.logger.debug("In Login View dispatch_request")
    user_id, id_token, auth_type, _ = auth.get_user_id_and_token()

    if not user_id:
      return self.json_error("NO USER_ID", 400)

    if not id_token:
      return self.json_error("NO_ID_TOKEN", 400)

    # Make auth object
    auth_ob = auth.Auth(id_token=id_token)

    # pragma pylint: disable=broad-except
    # Verify the id_token with Firebase
    try:
      claims = auth_ob.authorize_login_request(auth_type=auth_type)
    except Exception as e:
      return self.json_error(
          "UNAUTHORIZED_TOKEN" + str(e), 401, "Unauthorised login attempt:")

    # Check if id_token has been authorised
    # if claims is None, Unauthorised
    if not claims:
      return self.json_error("EMPTY_CLAIM", 401, "Unauthorised login attempt:")

    g.app.logger.info("Signing in with claims %s" % str(claims))

    if request.json:
      device_token_data = request.json.get('device_token', {})
    else:
      device_token_data = {}

    user = check_for_new_user(auth_ob.user_id, claims, auth_type, device_token_data, g.app.logger)

    session["gae_uid"] = user_id
    session["gae_token"] = id_token
    
    auth.login_callback(user)
    if auth_type == 'firebase':
      provider = claims.get('firebase').get('sign_in_provider')
    elif auth_type == 'auth0':
      provider = 'apple' if 'apple' in user.uid else 'google'


    response_dict = {"id": user.key().id,
                     "uId": user.uid,
                     "email": user.email,
                     "name": user.name,
                     "picture": user.picture,
                     "provider": provider,
                     }
    return self.json_success(response_dict, 202)



def check_for_new_user(user_id, claims, auth_type, device_token_data, logger):
  """
      checks if user with email exists in datastore.
      If exists, return games_played, games_won,
      Else, add to datastore
      with attributes
      user_id, email, games_played, games_won, join_date
  """

  logger.info("Checking if " + user_id + " is a new user")

  uid_filter = ('uid', '=', user_id)

  try:
    user = User.retrieve(filters=[uid_filter])[0]
  except IndexError:
    user = None
  except Exception as e:
    logger.info("Error trying to retrieve User %s", e)

  if not user:
    # Checking for user being an old cryptic cup member
    # from migration. This is not code that should be in
    # the library long term
    email_filter = ('email', '=', claims['email'])
    filters = [email_filter]

    ######
    #TODO: Handle multiple users
    ####
    try:
      user = User.retrieve(filters=filters)[0]
    except IndexError:
      user = None

  name = claims.get('name', '')
  if not name:
    email = claims['email']
    name = email[:email.find('@')]

  picture = claims.get('picture', get_app_or_default_prop('DEFAULT_PROFILE_IMAGE'))
  if user:
    user.update(
        uid=user_id,
        picture=picture,
        name=name,
        role=UserRole.DEFAULT.value
    )
  else:
    logger.info("Creating new user")
    logger.debug("\tUser ID: " + user_id)
    user = User(
        uid=user_id,
        email=claims['email'],
        picture=picture,
        name=name,
        role=UserRole.DEFAULT.value
    )
  if device_token_data:
    user.update(
      os=device_token_data.get('os'),
      device_token=device_token_data.get('token')
      )

  user.put()
  return user




"""
    This module defines a base Model class to create
    database models and upload entities to the datastore.
"""
from google.cloud import datastore
from .query import Query

from gaelib.db import properties, helpers


class Model():
  """
      The Model class to create
      database models and upload entities to the datastore.
  """
  __client__ = None

  def __init__(self, key=None, key_str='', **kwargs):
    object.__setattr__(self, "__entity_key__", None)
    object.__setattr__(self, "__entity__", None)
    object.__setattr__(self, "__retrieved__", False)
    client = self.get_client()
    """
            1. First we check for a passed key and set that as the key
            2. Then we check for a passed key str and use that to build
               a key which we set as a key
            3. Then if entity is passed, we set that passed entity as
               the entity. This is to be done only if we are not looking
               for the most recently fetched value of entity. Staleness
               must be expected
            4. Then we try to fetch the entity based on the set key
            5. Only if we see that there was no entity fetched, do we use
               allocate_ids to generate a key. This way we don't unnecessarily
               waste an rpc call on a key we are sure doesn't exist in the DS.
        """
    # Setting a key provided in init
    if key:
      self.__entity_key__ = key
    elif key_str:
      self.__entity_key__ = self.generate_key(key_str)

    # Setting entity either explicitly provided
    # or provided as a key/key_str
    entity = kwargs.pop('entity', None)

    if entity:
      self.__entity__ = entity
    elif self.__entity_key__:
      # Fetch only if the key or key_str were provided to init
      self.__entity__ = client.get(self.__entity_key__)
    else:
      # Only create the entity key if it was not already created
      # base on key str. This will not be dependent on whether the
      # entity was fetched or not.
      incomplete_key = self.get_client().key(
          (object.__getattribute__(self, '__class__').__name__))
      self.__entity_key__ = client.allocate_ids(incomplete_key, 1)[0]

    if self.__entity__:
      # This means that the datastore already had the entity for this key
      self.__retrieved__ = True
    else:
      self.__entity__ = datastore.Entity(key=self.__entity_key__)
      # Setting the default values for properties for absolutely
      # fresh entity
      all_attributes = dir(self)
      for attr in all_attributes:
        attribute = object.__getattribute__(self, attr)
        if isinstance(attribute, properties.Property):
          if attribute._default is not None:
            self.__entity__[attr] = attribute._default

    # kwargs items will overwrite those that were already in the entity
    for key, val in kwargs.items():
      attribute = object.__getattribute__(self, key)
      if isinstance(attribute, properties.Property):
        self.__entity__[key] = attribute.validate(val)

  def __getattribute__(self, attr):
    """
        The modified __getattr__ function to view the value of
        of the attributes in the model.
    """
    attribute = super().__getattribute__(attr)
    if isinstance(attribute, properties.Property):
      return self.__entity__.get(attr, None)
    return attribute

  def update(self, **kwargs):
    """
        The function to update multiple attributes of an entity at once.
    """
    for key, val in kwargs.items():
      attribute = object.__getattribute__(self, key)
      if isinstance(attribute, properties.Property):
        self.__entity__[key] = attribute.validate(val)

  def key(self):
    """
        Method to return cloud datastore key of entity.
    """
    return self.__entity_key__

  @classmethod
  def generate_key(cls, key_str):
    """
        The function to set the key for the entity.
    """
    return cls.get_client().key(cls.__name__, key_str)

  def put(self):
    """
        The function to insert the entity in cloud datastore
    """
    client = self.get_client()
    client.put(self.__entity__)

  @classmethod
  def get(cls, key):
    client = cls.get_client()
    entity = client.get(key)
    return cls(key=entity.key, entity=entity)

  @classmethod
  def retrieve(cls, filters=None, order=None, limit=None, key_strs=None,
               start_cursor=None):
    """
        The function to get the entity in cloud datastore
    """
    if key_strs:
      # Special case to get a multi get
      keys = [cls.generate_key(key_str) for key_str in key_strs]
      client = cls.get_client()
      entities = client.get_multi(keys)
      entity_list = [cls(key=entity.key, entity=entity) for entity in entities]
      ordered_entity_list = []
      for key in keys:
        for entity in entity_list:
          if entity.key() == key:
            ordered_entity_list.append(entity)
      return ordered_entity_list

    if not filters:
      filters = []
    query = Query(cls)

    if order:
      query.assign_order(order=order)

    for filter in filters:
      query.add_filter(filter[0], filter[1], filter[2])

    entities = query.fetch(limit=limit)
    return entities

  def delete(self):
    """
        The function to delete the entity from cloud datastore
    """
    client = self.get_client()
    client.delete(self.__entity_key__)

  def retrieved(self):
    return self.__retrieved__

  @classmethod
  def query(cls, **kwargs):
    """
        The method to create a Query object for the specified kind.
    """
    return Query(cls, **kwargs)

  @classmethod
  def get_client(self):
    """
        The method to get the datastore Client.
    """
    if self.__client__ is None:
      self.__client__ = datastore.Client(
          namespace=helpers.get_datastore_namespace())
    return self.__client__





import flask

blueprint = flask.Blueprint('filters', __name__)

# using the decorator
@blueprint.app_template_filter()
def datetime_difference(date1, date2):
  return int(((date2 - date1).total_seconds())/60)



from gaelib.view.base_view import BaseHttpHandler

from flask import g, request


class LogApi(BaseHttpHandler):
  """
      The contest home view function to get the previous contest winner
  """

  def dispatch_request(self):
    data = request.get_json()
    if 'logs' in data:
      for log in data['logs']:
        # Removing the timestamp and loglevel
        # from the client. We can add timestamp later if needed
        message = log['msg'][32:]
        if log['level'] == 'info':
          g.app.logger.info(message)
        elif log['level'] == 'debug':
          g.app.logger.debug(message)

    return self.json_success({}, 200)


####################### TESTCASES ##########################

from gaelib.tests.base import BaseUnitTestCase
from .model import Test

#point to note about using unittest.TestCase
#https://docs.djangoproject.com/en/3.1/topics/testing/overview/
"""
    List of test cases written
    1. Passing an already set key behavior
    2. Passing a key str behavior
    3. Passing an entity as keyword behavior
    4. Passing no id and getting from allocate ids
    5. Checking default properties are working
    6. Checking keywords passed are working
    7. Checking validation of props works
    8. Checking update and put works
    9. Checking delete works
    10.Checking query works
"""
class ModelUnitTestCase(BaseUnitTestCase):
    
    longMessage = False

    def create_object(self,**kwargs):
        obj = Test(**kwargs)
        return obj

    def test_passing_an_already_set_key_behaviour(self):
        #adding a key in the db to check with
        key = Test.generate_key('set_key_behaviour')
        model = Test(key = key, string1 = 'value1')
        model.put()      

        #passing the same key again to see the behaviour
        model = Test(key = key)
        self.assertTrue(model.retrieved(),"Key not found") 
    
    def test_key_str_behaviour(self):
        model = Test(key_str = 'key_str_behaviour')
        self.assertIsNotNone(model.key(), "Failed to generate key from key_str")
    
    def test_entity_as_keyword_behaviour(self): 
        model = Test(entity = self.create_object(string1 = 'value1', string2 = 'value2'))
        self.assertIsNotNone(model.__entity__, "Failed to create entity")

    def test_no_id_behaviour(self):
        model = Test()
        self.assertIsNotNone(model.key(), "Failed to generate key")
    
    def test_check_default_properties(self):
        model = self.create_object()
        attributes = ['string1','string2', 'int1', 'float1', 'bool1']
        for attribute in attributes:
            self.assertIsNone(model.__getattribute__(attribute), "Default value not none for attribute: {}".format(attribute))

    def check_keywords_passed(self, **kwargs):
        model = Test(**kwargs)
        model.put()
        for key, val in kwargs.items():
            self.assertEqual(model.__getattribute__(key),val, "Failed for key-value pair {} : {}".format(key,val))
      
    def test_check_keywords_passed(self):
        self.check_keywords_passed(string1 = 'one', string2 = 'two')

    def test_check_validation_of_props(self):
        self.assertRaises(ValueError, self.create_object, string1 = '123', string2 = 123, int1 = 123.40, float1 = True, bool1 = 123)

    def test_check_update_and_put(self):
        #checking put
        key = Test.generate_key('put and update')
        self.clear_database()

        model = Test(key = key, string1 = 'test_set_1')
        model.put()
        #model.retrieved() won't work here because the above instance of model is for creating the entity
        #and in that case self.__retrieved__ is set to false
        #We need to create another instance of the model with the same key and
        #then check for model.retrieved()
        model = Test(key = key)
        self.assertTrue(model.retrieved(),"Put failed")

        #checking update
        model.update(string1 = 'test_set_2')
        model.put()

        self.assertEqual(model.__getattribute__('string1'), 'test_set_2', "Update Failed")
        
    def test_check_delete(self):
        key = Test.generate_key('delete')
        #creatng a record to be deleted
        model = Test(key = key, string1 = "This record will be deleted")
        model.put()
        model = Test(key = key)
        self.assertTrue(model.retrieved(), "Failed to create entity")
        model.delete()
        model = Test(key = key)
        self.assertFalse(model.retrieved(), "Failed to delete key")
        
    def test_query(self):
        #create records to be fetched
        model = Test(key_str = 'record1', string1 = '123', string2 = '456', int1 = 111)
        model.put()
        model = Test(key_str = 'record2', string1 = '123', string2 = '456', int1 = 113)
        model.put()
        model = Test(key_str = 'record3', string1 = '123', string2 = '456', int1 = 112)
        model.put()

        #fetch records
        model = Test()
        q = model.query()
        q.add_filter('string1', '=', '123')
        q.add_filter('int1', '>', 111)
        q.assign_order('-int1') #descending
        last_value = 200
        for obj in q.fetch():
            self.assertGreaterEqual(last_value,obj.int1, "Failed to order items in result")
            last_value = obj.int1
        self.assertEquals(len(q.fetch()), 2, "Failed to fetch. Fetched {} instead of 2 items".format(len(q.fetch())))

    def test_retrieve(self):
        model = Test(key_str = 'record1', string1 = '123', string2 = '456', int1 = 111)
        model.put()
        model = Test(key_str = 'record2', string1 = '123', string2 = '456', int1 = 112)
        model.put()
        model = Test(key_str = 'record3', string1 = '123', string2 = '456', int1 = 113)
        model.put()
        model = Test(key_str = 'record3', string1 = '123', string2 = '456', int1 = 114)
        model.put()

        model = Test()

        #testing for key_str        
        items = model.retrieve(key_strs = ['record1'])
        self.assertEquals(items[0].int1, 111, "Failed to retrieve using key_str")

        #testing for filters, order and limit
        limit = 2
        items = model.retrieve(filters=[('string1', '=', '123'), ('int1', '>', 111)], order = '-int1', limit = limit)
        last_value = 200
        for obj in items:
            self.assertGreater(obj.int1, 111, "Failed to apply filter")
            self.assertEquals(obj.string1,'123', "Failed to apply filter")
            self.assertGreaterEqual(last_value,obj.int1, "Failed to order items")
            last_value = obj.int1
        self.assertEquals(len(items),limit, "Failed to apply limit. Fetched {} instead of {} items".format(len(items),limit))



from gaelib.tests.base import BaseUnitTestCase
from gaelib.cron.decorators import cron_validate
from werkzeug.exceptions import HTTPException
from mock import Mock
from main import app


class DecoratorsTestCase(BaseUnitTestCase):

  def setUp(self):
    super().setUp()

  def check_success(self):
    app.preprocess_request()
    decorated_func = cron_validate(Mock(return_value="OK"))
    response = decorated_func()
    self.assertEqual(response, "OK")

  def check_exception(self):
    app.preprocess_request()
    decorated_func = cron_validate(Mock())
    self.assertRaises(HTTPException, decorated_func)

  def test_cron_validate_when_ip_is_not_valid(self):
    environ_dict = {
        'REMOTE_ADDR': '0.1.0.2'
    }
    with app.test_request_context(environ_base=environ_dict):
      self.check_exception()

  def test_cron_validate_when_ip_is_valid(self):
    environ_dict = {
        'REMOTE_ADDR': '0.1.0.1'
    }
    with app.test_request_context(environ_base=environ_dict):
      self.check_success()

  def test_cron_validate_with_http_valid_ip(self):
    environ_dict = {
        'HTTP_X_FORWARDED_FOR': '0.1.0.1,x.x.x.x'
    }
    with app.test_request_context(environ_base=environ_dict):
      self.check_success()

  def test_cron_validate_with_http_forwarded_invalid_ip(self):
    environ_dict = {
        'HTTP_X_FORWARDED_FOR': '0.1.0.2,x.x.x.x'
    }
    with app.test_request_context(environ_base=environ_dict):
      self.check_exception()

  def test_cron_validate_with_appengine_header_and_valid_ip(self):
    environ_dict = {
        'REMOTE_ADDR': '0.1.0.1',
        'HTTP_X_AppEngine_Cron': ''
    }
    with app.test_request_context(environ_base=environ_dict, headers={}):
      self.check_success()

  def test_cron_validate_with_appengine_header_and_invalid_ip(self):
    environ_dict = {
        'REMOTE_ADDR': '0.1.0.1',
        'HTTP_X_AppEngine_Cron': ''
    }
    with app.test_request_context(environ_base=environ_dict, headers={}):
      self.check_success()



import os
from flask import g, jsonify, redirect, request, session

import constants
import gaelib.defaults as defaults


def is_dev():
  if os.getenv('GAE_DEPLOYMENT_ID', ''):
    return False
  else:
    return True

def is_prod():
  return not is_dev() and not is_staging()

def is_staging():
  return not is_dev() and 'staging' in os.getenv('GOOGLE_CLOUD_PROJECT')

def get_app_or_default_prop(prop_name):
  try:
    prop_value = getattr(constants, prop_name)
  except AttributeError:
    prop_value = ''
  if not prop_value:
    prop_value = getattr(defaults, prop_name)

  return prop_value

def get_dashboard_url_prefix():
  return get_app_or_default_prop('DASHBOARD_URL_PREFIX')

def get_post_login_page():
  return get_app_or_default_prop('POST_LOGIN_PAGE')  

def get_sidebar_template():
  return get_app_or_default_prop('SIDEBAR_TEMPLATE')

def dashboard_login_page():
  return redirect("/" + get_dashboard_url_prefix() + '/login', code=302)

def get_env():
  env = 'staging' if not is_prod() else 'prod'
  return env



import constants
from flask import g, jsonify, request, session
from gaelib.auth import auth
from gaelib.auth.models import User, UserRole
from gaelib.env import dashboard_login_page



def auth_required(view_function):
  """
      Function to be used with as_view().
      Returns a decorator function which authenticates a request
      before the target view's dispatch_request() is invoked.
  """
  def decorator(*args, **kwargs):
    """
        Authenticates request before invoking dispatch_request
        of target view.
    """
    response_dict = {"success": 0}

    user_id, id_token, auth_type, is_dashboard_url = auth.get_user_id_and_token()

    if is_dashboard_url and not (user_id and id_token):
      return dashboard_login_page()

    if not user_id:
      response_dict["error_message"] = "UNAUTHORIZED USER"
      return jsonify(response_dict), 401

    if not id_token:
      response_dict["error_message"] = "UNAUTHORIZED TOKEN"
      return jsonify(response_dict), 401

    auth_ob = auth.Auth(id_token, user_id)
    error = ''
    # pragma pylint: disable=broad-except
    try:
      auth_result = auth_ob.authorize_request(auth_type=auth_type)
    except Exception as e:
      error = e
      auth_result = None

    if not auth_result:
      error = "Unauthorised login for web request found"
      if is_dashboard_url:
        g.app.logger.warning(error)
        session.pop("gae_uid", None)
        return dashboard_login_page()

      response_dict["error_message"] = str(error)
      return jsonify(response_dict), 401

    uid_filter = ('uid', '=', user_id)
    user = User.retrieve(filters=[uid_filter])[0]
    g.user_key = user.key().id
    return view_function(*args, **kwargs)

  # The following if block is required to avoid different
  # view functions getting mapped to the same endpoint name
  # when the view_function is a LazyView object.
  # We explicitly change the decorator name to the LazyView object's
  # view_function name.
  # This ensures there's no conflicting view function to end point mapping.
  # Note that if this is not done, then registering more than 1 url rules
  # to the same blueprint (say solo) will get the same endpoint name (solo.decorator)
  # which raises an exception.
  # Refer to
  # https://stackoverflow.com/questions/17256602/assertionerror-view-function-mapping-is-overwriting-an-existing-endpoint-functi
  # for further details
  if view_function.__class__.__name__ == "LazyView":
    decorator.__name__ = view_function.view_function
  return decorator


def access_control(role):
  def decorator_without_arg(view_func):
    def access_control_decorator(*args, **kwargs):
      response_dict = {"success": 0}
      user_id, _, _, _ = auth.get_user_id_and_token()

      uid_filter = ('uid', '=', user_id)

      try:
        user = User.retrieve(filters=[uid_filter])[0]
      except IndexError:
        user = None

      if user:
        if UserRole(user.role) is role:
          return view_func(*args, **kwargs)
        else:
          g.app.logger.error(
              "User %s does not have access to staff only ", str(user.__dict__))
      else:
        g.app.logger.error("Unauthorised access, user doesn't exist")

      response_dict["error_message"] = "UNAUTHORIZED ACCESS"
      return jsonify(response_dict), 200
    return access_control_decorator

  return decorator_without_arg



from google.cloud import datastore
from . import helpers


def put_multi(entity_list):
  client = datastore.Client(namespace=helpers.get_datastore_namespace())
  client.put_multi(entity_list)
  return



"""
    This module contains all the url rules for the app.
"""

# pragma pylint: disable=invalid-name
import os
from flask import Blueprint

from gaelib.env import get_dashboard_url_prefix
from gaelib.view.base_view import LazyView


auth_urls = Blueprint('auth', __name__)
auth_urls.add_url_rule(
    '/login/',
    view_func=LazyView('gaelib.auth.views.Login', 'login'),
    methods=['POST'])

dashboard_lib_template_dir = os.path.abspath(
  './lib/dashboard/templates')
dashboard_lib_urls = Blueprint('dashboard_lib', __name__,
                               template_folder=dashboard_lib_template_dir)

dashboard_prefix = get_dashboard_url_prefix()
dashboard_lib_urls.add_url_rule(
  '/' + dashboard_prefix + '/login',
  view_func=LazyView(
    'gaelib.dashboard.views.DashboardLogin',
    'dashboard_login'),
  methods=['GET']
)
dashboard_lib_urls.add_url_rule(
  '/' + dashboard_prefix + '/logout',
  view_func=LazyView(
    'gaelib.dashboard.views.DashboardLogout',
    'dashboard_logout'),
    methods=['GET']
)
dashboard_lib_urls.add_url_rule(
  '/' + dashboard_prefix + '/notifications',
  view_func=LazyView(
    'gaelib.dashboard.views.Notifications',
    'notifications'),
    methods=['GET', 'POST']
)
# dashboard_lib_urls.add_url_rule(
#   '/' + dashboard_prefix + '/logout',
#   view_func=LazyView(
#     'gaelib.auth.views.DashboardLogout',
#     'dashboard_logout'),
#     methods=['GET']
# )


client_logger_urls = Blueprint('clientlogger', __name__)
client_logger_urls.add_url_rule(
  '/clientlogger',
  view_func=LazyView(
    'gaelib.clientlogger.logging.LogApi',
    'logger'),
  methods=['POST']
)



import unittest
from gaelib.db import helpers
from google.cloud import datastore
from mock import patch
from main import app


class BaseUnitTestCase(unittest.TestCase):
  def setUp(self):
    app.testing = True
    self.client = app.test_client()
    self.clear_database()

  def tearDown(self):
    self.clear_database()

  def clear_database(self):
    # Nosetests use a different database and it also does not use namespace from app.yaml,
    # so for now, we need to use the default namespace,
    # we will change this stuff later after further research.
    client = datastore.Client(namespace=helpers.get_datastore_namespace())
    query = client.query(kind='__kind__')
    query.keys_only()
    kinds = [entity.key.id_or_name for entity in query.fetch()]

    for kind in kinds:
      query = client.query(kind=kind)
      query.keys_only()
      keys = [entity.key for entity in query.fetch()]
      for i in range(0, len(keys), 499):
        client.delete_multi(keys[i:i + 499])

  def mock_storage_client(self):
    self.storage_client_patch = patch(
        'google.cloud.storage.Client')
    self.storage_client = self.storage_client_patch.start()

  def mock_bucket(self):
    self.bucket_patch = patch(
        'google.cloud.storage.Bucket')
    self.bucket = self.bucket_patch.start()

  def mock_blob(self):
    self.blob_patch = patch(
        'google.cloud.storage.Blob')
    self.blob = self.blob_patch.start()

  def mock_flask_request(self):
    self.request_patch = patch('flask.request')
    self.request = self.request_patch.start()



from flask import request, abort
from gaelib.cron import constants


def cron_validate(handler):
  def check_if_cron(*args, **kwargs):
    if request.environ.get('HTTP_X_FORWARDED_FOR') is None:
      client_ip = request.environ['REMOTE_ADDR']
    else:
      client_ip = request.environ['HTTP_X_FORWARDED_FOR'].split(',')[0]

    if (request.headers.get('X-AppEngine-Cron') is None) and (client_ip not in constants.VALID_CRON_JOB_IP_ADDRS):
      return abort(400)
    else:
      return handler(*args, **kwargs)
  return check_if_cron



from gaelib.tests.base import BaseUnitTestCase
from gaelib.db import helpers


class HelpersTestCase(BaseUnitTestCase):

  def setUp(self):
    super().setUp()

  def test_datastore_namespace(self):
    self.assertEqual(helpers.get_datastore_namespace(), 'DEFAULT')



from enum import Enum
from gaelib.db import model, properties


class UserRole(Enum):
  """
      Enum for User.role
  """
  DEFAULT = 0
  STAFF = 1
  ADMIN = 2


USER_ROLE_CHOICES = [user_role.value for user_role in UserRole]


class User(model.Model):
  """
      The database model for User kind.
  """
  email = properties.StringProperty()
  uid = properties.StringProperty()
  name = properties.StringProperty()
  picture = properties.StringProperty()
  role = properties.IntegerProperty(choices=USER_ROLE_CHOICES)
  os = properties.StringProperty()
  device_notification_token = properties.StringProperty()



"""
    This Module defines the BaseView Class
    which inherits View and has logger object
"""
from flask import g, jsonify, request
from flask.views import View
from werkzeug.utils import import_string, cached_property

from gaelib.auth.decorators import auth_required
from gaelib.cron.decorators import cron_validate


class LazyView(object):
  """
      This class is used to import the view class
      when needed and prevent upfront import.
  """

  def __init__(self, import_name, view_function):
    self.__module__, self.__name__ = import_name.rsplit('.', 1)
    self.import_name = import_name
    self.view_function = view_function

  @cached_property
  def view(self):
    return import_string(self.import_name).as_view(self.view_function)

  def __call__(self, *args, **kwargs):
    return self.view(*args, **kwargs)


class BaseHttpHandler(View):
  """ BaseView Class """

  def json_response(self, response_body, status, headers=None):
    if not isinstance(response_body, dict):
      g.app.logger.error("Response body not a dict, cannot convert to JSON")
      raise TypeError(
          "'\nThe view function return type must be a dictionary,'"
          "'but it was a {response_body.__class__.__name__}.'"
          .format(response_body=response_body))

    response_body = jsonify(response_body)

    if headers is None:
      headers = {}
    headers["Content-Type"] = "application/json"
    return response_body, status, headers

  def json_error(self, error_message, status_code, warning_message=None):
    """
        Logs error messages, defines appropriate response_dict,
        and returns the same with status code.
    """
    if warning_message is not None:
      g.app.logger.warning(warning_message)

    g.app.logger.error(error_message)
    response_dict = {}
    response_dict["success"] = 0
    response_dict["error_message"] = error_message
    return self.json_response(response_dict, status_code)

  def json_success(self, response_dict, status_code):
    """
        Updates response_dict with success=1.
        Returns the same with status code.
    """
    response_dict["success"] = 1
    return self.json_response(response_dict, status_code)

  def validate_request_args(self):
    request_args = request.args.to_dict()
    self.missing_args = []

    for args in self.required_request_args:
      if args not in request_args:
        self.missing_args.append(args)

    if not self.missing_args:
      return True
    return False

  def validation_error(self):
    if self.missing_args:
      return self.json_error("Missing Arguements: " + str(self.missing_args), 200)


class BaseApiHandler(BaseHttpHandler):
  decorators = [auth_required]


class BaseCronJobHandler(BaseHttpHandler):
  decorators = [cron_validate]



"""
    This module contains all defaults available for when
    the application importing the lib doesnt set values
"""
DASHBOARD_URL_PREFIX = 'admindashboard'
POST_LOGIN_PAGE = 'users'
SIDEBAR_TEMPLATE = 'dashboard/operations_sidebar'


import json
import jwt
import time
from firebase_admin import credentials, initialize_app, messaging
from flask import g
from hyper import HTTPConnection

from gaelib.env import get_env, get_app_or_default_prop



ALGORITHM = 'ES256'

if get_env() == 'staging':
  APNS_KEY_ID = get_app_or_default_prop('STAGING_APNS_KEY_ID') # 'PVQ627R278'
  APNS_TEAM_ID = get_app_or_default_prop('STAGING_APNS_TEAM_ID') #'DH5NSA4U89'
  APNS_AUTH_KEY = get_app_or_default_prop('STAGING_APNS_AUTH_KEY')
  conn = HTTPConnection('api.sandbox.push.apple.com:443')
else:
  APNS_KEY_ID = get_app_or_default_prop('PROD_APNS_KEY_ID') # 'PVQ627R278'
  APNS_TEAM_ID = get_app_or_default_prop('PROD_APNS_TEAM_ID') #'DH5NSA4U89'
  APNS_AUTH_KEY = get_app_or_default_prop('PROD_APNS_AUTH_KEY')
  conn = HTTPConnection('api.sandbox.push.apple.com:443')

IOS_BUNDLE_ID = get_app_or_default_prop('IOS_BUNDLE_ID') # 'com.codejedi.crypticcup1'

f = open(APNS_AUTH_KEY)
secret = f.read()

def send_apns_notification(token_id, title, message):
  token = jwt.encode(
    {
        'iss': APNS_TEAM_ID,
        'iat': time.time()
    },
    secret,
    algorithm= ALGORITHM,
    headers={
        'alg': ALGORITHM,
        'kid': APNS_KEY_ID,
    }
  )
  request_headers = {
    'apns-expiration': '0',
    'apns-priority': '10',
    'apns-topic': BUNDLE_ID,
    'authorization': 'bearer {0}'.format(token.decode('ascii'))
  }
  payload = {
              'aps':{
                      'alert' : {
                                 'title' : title,
                                 'body' : message,
                                }
                     }
            }
  
  payload = json.dumps(payload).encode('utf-8')
  path = '/3/device/{0}'.format(token_id)
  conn.request(
    'POST',
    path,
    payload,
    headers=request_headers
  )
  resp = conn.get_response()
  print(resp.status)
  print(resp.read())

def send_fcm_notification(registration_tokens, title, message):
  cred = credentials.RefreshToken('<path to firebase credentials>.json')
      initialize_app(cred)
      registration_tokens = registration_tokens.
      message = messaging.MulticastMessage(
        data={'title': title, 'message': message},
        tokens=registration_tokens,
      )
  response = messaging.send_multicast(message)
  print(response.failure_count)



from setuptools import setup, find_packages

setup(
    name='gaelib',
    version='0.1.2',
    description='Google App Engine Library',
    author='Shantanu Mallik',
    license='MIT',
    packages=find_packages(where='.'),
    #package_dir={'gaelib':'./', 'clientlogger':'./gaelib/','dashboard':'./gaelib/','tests':'./gaelib/', 'db':'./gaelib/', 'view':'./gaelib/', 'storage':'./gaelib/', 'utils':'./gaelib/', 'auth':'./gaelib/'},
    zip_safe=False,
    install_requires=[
        'wheel',
    	'google-cloud-speech',
        'google-cloud-core',
        'grpcio',
        'google-auth',
        'google-cloud-datastore',
        'requests',
        'google-cloud-logging',
        'inflect==2.1.0',
        'google-cloud-storage',
        'py-dateutil==2.2',
        'cryptography==2.9.2',
        'pyjwt==1.7.1',
        'hyper==0.7.0',
        'google-cloud-tasks==1.5.0',
        'googleapis_common_protos',
        'firebase-admin',
        'python-jose',
        'nose',
        'mock',
        'google-cloud',
        'click==7.1.2',
        'Flask==1.1.2',
        'Flask-Cors==3.0.9',
        'gunicorn==20.0.4',
        'itsdangerous==1.1.0',
        'Jinja2==2.11.2',
        'MarkupSafe==1.1.1',
        'python-dotenv==0.14.0',
        'six==1.15.0',
        'Werkzeug==1.0.1',
        'constants'
    ],
    setup_requires=[
        'wheel'
    ]
)


"""
    This module defines a base Query class to create
    datastore query objects and perform queries on datastore.
"""
from google.cloud import datastore


class Query():
  """
      The class for creating a datastore query object
      and perform queries on a model.
  """

  def __init__(self, model_class, **kwargs):
    self.__model_class__ = model_class
    self.__query__ = model_class.get_client().query(
        kind=self.__model_class__.__name__, **kwargs)

  def add_filter(self, attribute, op, val):
    """
        The method to add filters for the query.
    """
    self.__query__.add_filter(attribute, op, val)

  def assign_order(self, order):
    self.__query__.order = order

  def fetch(self, **kwargs):
    """
        The method to fetch and return query results.
    """
    results = []
    for obj in list((self.__query__.fetch(**kwargs))):
      result_obj = self.__model_class__(key=obj.key, entity=obj)
      results.append(result_obj)
    return results



import os


def get_default_storage_bucket():
  return str(os.getenv('GOOGLE_CLOUD_PROJECT')) + '.appspot.com'


def get_storage_bucket():
  return os.getenv('CLOUD_STORAGE_BUCKET', default=get_default_storage_bucket())
